/**
***  Copyright (c) 1995, 1996, 1997, 1998, 1999, 2000 by
***  The Board of Trustees of the University of Illinois.
***  All rights reserved.
**/

/*****************************************************************************
 * $Source: /Projects/namd2/cvsroot/namd2/src/SimParameters.h,v $
 * $Author: jim $
 * $Date: 2012/03/11 23:40:50 $
 * $Revision: 1.1204 $
 *****************************************************************************/

#ifndef SIMPARAMETERS_H
#define SIMPARAMETERS_H

#include "common.h"
#include "Vector.h"
#include "Lattice.h"

#include "MGridforceParams.h"

class ParseOptions;
class Communicate;
class ConfigList;
class MIStream;

//  The class SimParameters is really just a glorified structure used to
//  maintain the global simulation parameters.  The only functions
//  associated with the class are used to get the parameters from the
//  ConfigList object, to send that Parameters from the master node 
//  to the other nodes, and to receive the Parameters on the other nodes.


//  The following definitions are used to distinguish between possible
//  bonded exclusion settings
typedef int  ExclusionSettings;

#define NONE		0
#define	ONETWO  	1
#define	ONETHREE	2
#define	ONEFOUR		3
#define	SCALED14 	4

//  The following definitions are used to distinguish between multiple
//  timestep integration schemes
typedef int  MTSChoices;

#define NAIVE		0
#define VERLETI		1

//  The following definitions are used to distinuish between multiple
//  long-short range force splittings
#define SHARP		0
#define XPLOR		1
#define C1		2
#define C2              3

//  The following definitions are used to distinguish among load
//  balancers and their strategies
#define LDBAL_NONE		0
#define LDBAL_CENTRALIZED	1	// default
#define LDBAL_HYBRID		2

#define LDBSTRAT_DEFAULT	10	// default
#define LDBSTRAT_COMPREHENSIVE	11
#define LDBSTRAT_REFINEONLY	12
#define LDBSTRAT_OLD		13

// The following definitions are used to distinguish between patch-splitting
// strategies
#define SPLIT_PATCH_POSITION	0	// atom position determines patch
#define SPLIT_PATCH_HYDROGEN	1	// hydrogen groups are not broken up

// The following definitions are used to distinguish the range of rigid
// bond calculations: none, all bonds to hydrogen, or only water
#define RIGID_NONE    0
#define RIGID_ALL     1
#define RIGID_WATER   2

// Added by JLai -- The following definitions are used to distinguish
// the different GoMethodologies available to the Go program
// -- 6.3.11
typedef int GoChoices;
#define GO_MATRIX 1
#define GO_SPARSE 2
#define GO_LOWMEM 3

class SimParameters
{
private:
public:

//  MAKE SURE THAT THIS CLASS CAN BE BIT COPIED OR YOU WILL HAVE TO
//  ADD SPECIAL CODE TO send_SimParameters() and receive_SimParameters()

  Bool lonepairs;  // enable lone pairs
  int watmodel; // integer code for the water model in use
                // choices are defined in common.h
  Bool LJcorrection; // flag for whether water tail corrections should be used
	BigReal dt;	   		//  Timestep size
	int N;		   		//  Number of steps to be performed
	int stepsPerCycle;		//  Number of timesteps per cycle

	zVector cellBasisVector1;	//  Basis vector for periodic cell
	zVector cellBasisVector2;	//  Basis vector for periodic cell
	zVector cellBasisVector3;	//  Basis vector for periodic cell
	zVector cellOrigin;		//  Fixed center of periodic cell
	Lattice lattice;		//  All data for periodic cell
	
	int nonbondedFrequency;		//  Number of timesteps between
					//  nonbonded evaluation
	int fullElectFrequency;		//  Number of timesteps between
					//  full electrostatic evaluation
        BigReal fmaTheta;	        //  DPMTA theta value
	int ldBalancer;			//  None, Centralized or Hybrid
	int ldbStrategy;                //  What load balancing strategy to use
	int ldbPeriod;                  //  How often to do load balancing
	int firstLdbStep;		//  What step to do the first 
                                        //  load-balance on.
	int lastLdbStep;		//  What step to do the last
                                        //  load-balance on.
        int hybridGroupSize;            //  hybrid group size
	BigReal ldbBackgroundScaling;	//  scaling factor for background load
	BigReal ldbPMEBackgroundScaling;//  scaling factor for PME background
	BigReal ldbHomeBackgroundScaling;//  scaling factor for home background
	BigReal ldbRelativeGrainsize;   //  fraction of average load per compute
	
	int traceStartStep; //the timestep when trace is turned on, default to 3*firstLdbStep;
	int numTraceSteps; //the number of timesteps that are traced, default to 2*ldbPeriod;
	
#ifdef MEASURE_NAMD_WITH_PAPI
	Bool papiMeasure; //default to false
	int papiMeasureStartStep; //the timestep when to measure using PAPI, default to 3*firstLdbStep;
	int numPapiMeasureSteps; //the number of timesteps when performance are measured with PAPI, default to 40;
#endif
	
	Bool outputMaps; //control whether to dump compute/patch map before load balancing
	Bool simulateInitialMapping; //if true, the initial mapping during startup is dumped and exit
	int simulatedPEs;
	int simulatedNodeSize;

	Bool benchTimestep; //only cares about benchmarking the timestep, so no file output to save SUs for large-scale benchmarking

	int twoAwayX;			//  half-size patches in X dimension
	int twoAwayY;			//  half-size patches in Y dimension
	int twoAwayZ;			//  half-size patches in Z dimension
	int maxPatches;			//  maximum patch count
	Bool ldbUnloadPME;		//  unload processors doing PME
	Bool ldbUnloadZero;		//  unload processor 0
	Bool ldbUnloadOne;		//  unload processor 1 
	Bool noPatchesOnZero;		//  no patches on processor 0
	Bool noPatchesOnOne;		//  no patches on processor 1
	
	BigReal initialTemp;   		//  Initial temperature for the 
					//  simulation
	Bool comMove;     		//  Should the center of mass be 
					//  able to move
	Bool zeroMomentum;		//  remove momentum drift from PME
	Bool zeroMomentumAlt;		//  alternate method for testing
	Bool wrapWater;			//  Wrap water around on output
	Bool wrapAll;			//  Wrap clusters around on output
	Bool wrapNearest;		//  Wrap to closest image to origin
	BigReal dielectric;   		//  Dielectric constant
	ExclusionSettings exclude;      //  What electrostatic exclusions should
					//  be made
	BigReal scale14;		//  Scaling factor for 1-4 
					//  electrostatics
	BigReal nonbondedScaling;	//  Scaling factor for nonbonded forces
	int dcdFrequency;		//  How often (in timesteps) should
					//  a DCD trajectory file be updated
  int dcdUnitCell;  // Whether to write unit cell information in the DCD
	int velDcdFrequency;		//  How often (in timesteps) should
					//  a velocity DCD file be updated
	int forceDcdFrequency;		//  How often (in timesteps) should
					//  a force DCD file be updated
	int xstFrequency;		//  How often (in timesteps) should
					//  a XST trajectory file be updated
	char auxFilename[128];		//  auxilary output filename
	char dcdFilename[128];		//  DCD filename
	char velDcdFilename[128];       //  Velocity DCD filename
	char forceDcdFilename[128];     //  Force DCD filename
	char xstFilename[128];		//  Extended system trajectory filename
	char outputFilename[128];	//  Output file name.  This name will
					//  have .coor appended to it 
					//  for the coordinates and 
					//  .vel appended to
					//  it for the velocities
	char restartFilename[128];	//  Base name of the restart file
	int restartFrequency;		//  How often (in timesteps) shoud the
					//  restart files be updated
        Bool restartSave;		//  unique filenames for restart files
	Bool binaryRestart;		//  should restart files be
					//  binary format rather than PDB
	Bool binaryOutput;		//  should output files be
					//  binary format rather than PDB
	BigReal cutoff;			//  Cutoff distance
	BigReal margin;			//  Fudge factor on patch size
	BigReal patchDimension;		//  Dimension of each side of a patch
					//  This is either cutoff+margin or
					//  pairlistDist+margin depending on
					//  whether or not switching is on
					//  or not
	BigReal limitDist;		//  Distance below which nonbonded
					//  forces between atoms are limited
	Bool switchingActive;		//  Flag TRUE->using switching function
					//  for electrostatics and vdw
	Bool vdwForceSwitching;		//  Flag TRUE->using force switching
					//  function for vdw
	BigReal switchingDist;		//  Distance at which switching
					//  becomes active
	Bool martiniSwitching;		//  Flag TRUE->use Martini residue-based
                                        //  coarse-grain switching function
	Bool martiniDielAllow;          //  Allow non-standard dielectric constant 
                                        //  for use with Martini when dielectric != 15.0
	BigReal pairlistDist;		//  Distance within which atom pairs 
					//  should be added to pairlist
	int pairlistMinProcs;		//  Minimum number of processors
					//  to enable pairlists
	int usePairlists;		//  Derived from pairlistMinProcs

	int pairlistsPerCycle;		//  regenerate x times per cycle
	BigReal pairlistShrink;		//  tol *= (1 - x) on regeneration
	BigReal pairlistGrow;		//  tol *= (1 + x) on trigger
	BigReal pairlistTrigger;	//  trigger is atom > (1 - x) * tol
	int outputPairlists;		//  print pairlist warnings this often

	Bool constraintsOn;		//  Flag TRUE-> harmonic constraints 
					//  active
	int constraintExp;		//  Exponent for harmonic constraints

	/* BEGIN gf */
	Bool gridforceOn;		//  Flag TRUE -> gridforce active
	Bool gridforceVolts;		//  Flag TRUE -> gridforce using volts as units
	zVector gridforceScale;		//  Gridforce scale factor
	Bool gridforceContA1;		//  Flag TRUE -> grid continuous in A1 direction
	Bool gridforceContA2;		//  Flag TRUE -> grid continuous in A2 direction
	Bool gridforceContA3;		//  Flag TRUE -> grid continuous in A3 direction
	zVector gridforceVOffset;	//  Gridforce potential offsets
	Bool gridforceLite;		//  Flag TRUE -> use lightweight, fast, feature-poor gridforce
	/* END gf */
        Bool mgridforceOn;
        MGridforceParamsList mgridforcelist;

        //****** BEGIN selective restraints (X,Y,Z) changes 
        Bool selectConstraintsOn;       //  Flag TRUE-> selective restraints  
                                        //  active
        Bool constrXOn, constrYOn,       
             constrZOn;                 //  Flag TRUE-> select which Cartesian 
                                        //  component to restrain
        //****** END selective restraints (X,Y,Z) changes 
	BigReal constraintScaling;	//  Scaling factor for constraint forces

        //****** BEGIN CHARMM/XPLOR type changes
        Bool paraTypeXplorOn;           //  FLAG TRUE-> parametrs are XPLOR format (default)
        Bool paraTypeCharmmOn;          //  FLAG TRUE-> parametrs are CHARMM format
        //****** END CHARMM/XPLOR type changes

	// Ported by JLai -- JE - Go
        Bool goForcesOn;          //  FLAG TRUE-> Go forces will be calculated
        char goParameters[128];   //  File for Go parameters
        char goCoordinates[128];  //  File for Go structure and atom chain types
        //JLai 6.3.11
	GoChoices  goMethod;      //  Integer for Go method -- 1) Matrix-Go, 3) Low-mem-Go
	// End of port -- JL

        //****** BEGIN moving constraints changes 
        Bool movingConstraintsOn;       //  Flag TRUE-> moving constraints 
                                        //  active
        zVector movingConsVel;           //  Velocity of the movement, A/timestep
        //****** END moving constraints changes 
        //****** BEGIN rotating constraints changes 
        Bool rotConstraintsOn;          //  Flag TRUE-> rotating constraints 
                                        //  active
        zVector rotConsAxis;             //  Axis of rotation
        zVector rotConsPivot;            //  Pivot point of rotation
        BigReal rotConsVel;             //  Velocity of rotation, Deg/timestep
        //****** END rotating constraints changes 

        //****** BEGIN moving drag changes
        Bool movDragOn;               //  Flag TRUE-> moving drag active
        char movDragFile[128];        //  PDB file defining dragged atoms
                                      //  by non-zero value in the column
	BigReal movDragGlobVel;       //  global drag velocity (A/step)
	char movDragVelFile[128];     //  PDB file; XYZ scale moving drag
                                      //  velocity for each atom
        //****** END moving drag changes
        //****** BEGIN rotating drag changes
        Bool rotDragOn;               //  Flag TRUE-> rotating drag active
        char rotDragFile[128];        //  PDB file defining dragged atoms
                                      //  by non-zero value in the column
	char rotDragAxisFile[128];    //  PDB file; XYZ define axes for atoms;
	char rotDragPivotFile[128];   //  PDB file; XYZ define pivots for atoms
	BigReal rotDragGlobVel;       //  global drag velocity (deg/step)
	char rotDragVelFile[128];     //  PDB file; B or O scales angular
                                      //  velocity for each atom
        //****** END rotating drag changes
        //****** BEGIN "constant" torque changes
        Bool consTorqueOn;            //  Flag TRUE-> "constant" torque active
        char consTorqueFile[128];     //  PDB file defining torqued atoms
                                      //  by non-zero value in the column
	char consTorqueAxisFile[128]; //  PDB file; XYZ define axes for atoms;
	char consTorquePivotFile[128];//  PDB file; XYZ define pivots for atoms
	BigReal consTorqueGlobVal;    //  global "torque" (Kcal/(mol*A^2))
	char consTorqueValFile[128];  //  PDB file; B or O scales "torque"
                                      //  for each atom
        //****** END "constant" torque changes

        //****** BEGIN SMD constraints changes   
        Bool SMDOn;                     //  Flag TRUE-> SMD constraints active
        BigReal SMDVel;                 //  Velocity of the movement, A/timestep
        zVector SMDDir;                  //  Direction of the movement
        BigReal SMDk; 			//  Elastic constant for SMD
	BigReal SMDk2;			//  Transverse elastic constant for SMD
 	char SMDFile[128];		//  File for SMD information
        int SMDOutputFreq;              //  Output frequency for SMD constr.
        //****** END SMD constraints changes 
        
  //****** BEGIN tabulated energy section
  Bool tabulatedEnergies;
  int tableNumTypes;
  char tabulatedEnergiesFile[128];
  char tableInterpType[128];
  Real tableSpacing;
  BigReal tableMaxDist;
  //****** END tabulated energy section

        // TMD
        Bool TMDOn, TMDDiffRMSD;
        BigReal TMDk;
        char TMDFile[128], TMDFile2[128];
        int TMDOutputFreq;
        int TMDFirstStep, TMDLastStep;
        BigReal TMDInitialRMSD, TMDFinalRMSD;

        //Symmetry restraints
        Bool symmetryOn, symmetryScaleForces;
        BigReal symmetryk;
        char symmetrykfile[128];
        char symmetryFile[128];
        char symmetryMatrixFile[128];
        int symmetryFirstStep, symmetryLastStep, symmetryFirstFullStep, symmetryLastFullStep;

        
// Modifications for alchemical simulations
// Begin alch flags

  Bool alchOn;              //  Doing alchemical simulation?
  Bool alchFepOn;           //  Doing alchemical simulation?
  Bool alchThermIntOn;      //  Doing thermodynamic integration?
  Bool alchFepWCARepuOn;    //  Doing WCA decomposition repulsion interaction?
  Bool alchFepWCADispOn;    //  Doing WCA decomposition dispersion interaction?
  int alchMethod;           //  Which alchemical method to use? fep or ti
  BigReal alchLambda;       //  lambda for dynamics
  BigReal alchLambda2;      //  lambda for comparison
  BigReal alchFepWCArcut1;  //  rcut1 of WCA decompistion repulsion
  BigReal alchFepWCArcut2;  //  rcut2 of WCA decomposition repulsion
  BigReal alchTemp;         //  temperature for alchemical calculation
  int alchOutFreq;          //  freq. of alchemical output
  Bool alchEnsembleAvg;      //if do ensemble average for the net free energy difference 
  char alchOutFile[128];    //  alchemical output filename
  int alchEquilSteps;       //  # of equil. steps in the window
  BigReal alchVdwShiftCoeff; //  r2 shift coeff used for generating  
                            //  the alchemical altered vdW interactions
  BigReal alchElecLambdaStart;  //  lambda value for starting point of
                                //  electrostatic interactions of 
                                //  exnihilated particles.  For annihilated
                                //  particles the starting point is
                                //  (1-alchElecLambdaStart)
  BigReal alchVdwLambdaEnd;  //  lambda value for endpoint of vdW
                             //  interactions of exnihilated particles.
                             //  For annihilated particles the endpoint is 
                             //  (1-alchVdwLambdaEnd)
  Bool alchDecouple;  // alchemical decoupling rather than annihilation

// End alch flags
//fepe

  
	Bool lesOn;			//  Locally enhanced sampling?
	int lesFactor;			//  local enhancement factor
	Bool lesReduceTemp;		//  Reduce enhanced atom temperature?
	Bool lesReduceMass;		//  Reduce enhanced atom mass?

        Bool extForcesOn;		//  Are ext command forces present?
        char extForcesCommand[256];
        char extCoordFilename[128];
        char extForceFilename[128];

  Bool printBadContacts;        //print indices of bad contacts being moved downhill

  //gbis implicit solvent parameters
  Bool GBISOn;                    //do generalized born implicit solvent
  BigReal fsMax;
  Bool GBISserOn;                 //do generalized born implicit solvent serial
  BigReal solvent_dielectric;     //epsilon_s
  BigReal coulomb_radius_offset;  //rho_0
  BigReal kappa;      //debye screening length; k = sqrt(ion concentration mol/L ) / 0.304
  BigReal ion_concentration;
	BigReal gbis_delta;							//three parameters for born radius calc
	BigReal gbis_beta;
	BigReal gbis_gamma;
	BigReal alpha_cutoff;						//pairwise cutoff for integrating born radius
	BigReal alpha_max;								//maximum allowable born radius
  Bool LCPOOn;                    //do LCPO SASA for GBSA
  BigReal surface_tension;        //surface tension (kcal/mol/Ang^2) for LCPO

        Bool drudeOn;       // Perform integration of Drude oscillators?
        BigReal drudeTemp;  // (low) temperature for freezing Drude oscillators
        BigReal drudeDamping;    // Langevin damping coefficient (1/ps)
                                 //   defaults to langevinDamping
        BigReal drudeBondLen;    // Length beyond which to apply quartic
                                 //   restraining potential to Drude bond
        BigReal drudeBondConst;  // Force constant for restraining potential
	BigReal drudeNbtholeCut;             // Radius of thole pair interaction

	Bool pairInteractionOn;		//  Calculate pair interactions?
	int pairInteractionGroup1;	//  Interaction group 1.
	int pairInteractionGroup2;	//  Interaction group 2.
        Bool pairInteractionSelf;       //  Compute just within group.
     
        Bool cosAngles;    // Can some angles be cos-based
	Bool globalForcesOn;		//  Are global forces present?
	Bool tclForcesOn;		//  Are Tcl forces present?
#ifdef NAMD_TCL
        Bool tclIsThreaded;             //  Is Tcl library thread-safe?
#endif
	Bool tclBCOn;			//  Are Tcl boundary forces present
	char *tclBCScript;		//  Script defining tclBC calcforces
	char tclBCArgs[128];		//  Extra args for calcforces command
	Bool freeEnergyOn;		//  Doing free energy perturbation?
	Bool miscForcesOn;		//  Using misc forces?
	Bool colvarsOn;         //  Using the colvars module?

	Bool fixedAtomsOn;		//  Are there fixed atoms?
	Bool fixedAtomsForces;		//  Calculate forces anyway?

	Bool langevinOn;		//  Flag TRUE-> langevin dynamics active
	BigReal langevinTemp;		//  Temperature for Langevin dynamics
	BigReal langevinDamping;	//  Damping coefficient (1/ps)
	Bool langevinHydrogen;		//  Flag TRUE-> apply to hydrogens
	
	// BEGIN LA
	Bool loweAndersenOn;		//  Flag TRUE-> Lowe-Andersen dynamics active
	BigReal loweAndersenTemp;	//  Temperature for Lowe-Andersen dynamics
	BigReal loweAndersenRate;	//  Collision frequency for Lowe-Andersen dynamics (1/ps)
	BigReal loweAndersenCutoff;	//  Cutoff radius for Lowe-Andersen dynamics
	// END LA

	Bool globalOn;			//  Flag TRUE-> use global integrator
	Bool dihedralOn;		//  Flag TRUE-> dihedral dynamics active
	Bool COLDOn;			//  Flag TRUE-> constrained overdamped
					//  langevin dynamics active
	BigReal COLDRate;		//  Damping coefficient for COLD.
	BigReal COLDTemp;		//  Temperature for COLD.

	Bool tCoupleOn;			//  Flag TRUE-> Temperature coupling 
					//  active
	BigReal tCoupleTemp;		//  Temperature for temp coupling

	int rescaleFreq;		//  Velocity rescale frequency
	BigReal rescaleTemp;		//  Temperature to rescale to

        Bool accelMDOn;                 //  Perform accelerated MD
        Bool accelMDdihe;               //  Apply boost to the dihedral potential
        Bool accelMDdual;               //  dual boost mode  
        Bool accelMDDebugOn;            //  Debugging accelerated MD
        BigReal accelMDFirstStep;       //  First aMD step
        BigReal accelMDLastStep;        //  Last aMD step
        int accelMDOutFreq;             //  aMD output frequency
        BigReal accelMDE;               //  aMD E
        BigReal accelMDalpha;           //  aMD alpha
        BigReal accelMDTE;              //  E for total potential in the dual boost mode
        BigReal accelMDTalpha;          //  alpha for total potential in the dual boost mode

        /* Begin Adaptive Temperature Sampling */
        Bool adaptTempOn;                      //  is adaptTempOn
        Bool adaptTempDebug;                   //  Debuggin adaptive temperature sampling
        int adaptTempFirstStep;                //  First adaptTemp step
        int adaptTempLastStep;                 //  Last adaptTemp step
        int adaptTempOutFreq;                  //  adaptTemp output frequency
        int adaptTempFreq;                     //  Steps between adaptTemp updates
        BigReal adaptTempTmin;                 //  Lower temperature bound
        BigReal adaptTempTmax;                 //  Upper temperature bound
        BigReal adaptTempAutoDt;               //  Auto jump size. Value determines upper bound, adaotTempDt determines lower bound 
        int adaptTempBins;                     //  Number of bins to store average energy values
        BigReal adaptTempDt;                   //  timestep for adaptTemp updates - only affects Temperature random walk
        BigReal adaptTempCgamma;               //  Cgamma variable for adaptive bin averaging Cgamma = 0 is normal Averaging. 1 > Cgamma >= 0
        Bool adaptTempLangevin;                //  Couple to Langevin Thermostat
        Bool adaptTempRescale;                 //  Couple to Vel. Rescaling
        char adaptTempInFile[128];             //  Restart information for adaptTemp to read
        char adaptTempRestartFile[128];        //  File to write restart information
        int  adaptTempRestartFreq;             //  Frequency of writing restart output
        Bool adaptTempRandom;                  //  Do we assign random temperatures when we step out of [Tmin,Tmax]?
        /* End Adaptive Temperature Sampling */

	int reassignFreq;		//  Velocity reassignment frequency
	BigReal reassignTemp;		//  Temperature to reassign to
	BigReal reassignIncr;		//  Added to reassignTemp each time
	BigReal reassignHold;		//  Hold reassignTemp at this value

	Bool useGroupPressure;		//  Use group rather than atomic
					//  quantities for pressure calc

	Bool excludeFromPressure;	//  Flag TRUE-> some atoms not rescaled
 
	Bool useFlexibleCell;		//  Use anisotropic cell fluctuations
	Bool useConstantArea;		//  x,y dimensions fixed.
	Bool useConstantRatio;		//  x,y ratio fixed.

  Bool fixCellDims; // fix the cell dimensions
  Bool fixCellDimX;
  Bool fixCellDimY;
  Bool fixCellDimZ;

	Bool berendsenPressureOn;	//  Berendsen pressure bath
	BigReal berendsenPressureTarget;
	BigReal berendsenPressureCompressibility;
	BigReal berendsenPressureRelaxationTime;
	int berendsenPressureFreq;

	Bool langevinPistonOn;		//  Langevin piston pressure control
	BigReal langevinPistonTarget;
	BigReal langevinPistonPeriod;
	BigReal langevinPistonDecay;
	BigReal langevinPistonTemp;

        BigReal surfaceTensionTarget;

        Bool pressureProfileOn;         // Compute lateral pressure profile?
        int pressureProfileSlabs;       // Number of slabs
        int pressureProfileFreq;        // How often to store profile data
        int pressureProfileAtomTypes;
        Bool pressureProfileEwaldOn;    // Compute Ewald contribution?
        int pressureProfileEwaldX;
        int pressureProfileEwaldY;
        int pressureProfileEwaldZ;
        
	zVector strainRate;
	zVector strainRate2; // off diagonal elements (xy, xz, yz)

	unsigned int randomSeed;	//  Seed for random number generator

	Bool FMAOn;                     //  Flag TRUE-> FMA active
	int FMALevels;			//  Number of Levels for FMA
	int FMAMp;			//  Number of multipole terms for FMA
	Bool FMAFFTOn;			//  FFT on/off flag for FMA
	int FMAFFTBlock;		//  FFT blocking factor for FMA

	Bool fullDirectOn;		//  Should direct calculations of
					//  full electrostatics be performed?

        Bool MSMOn;
        int MSMApprox;
        int MSMSplit;
        int MSMLevels;
        int MSMBlockSizeX;
        int MSMBlockSizeY;
        int MSMBlockSizeZ;
        BigReal MSMGridSpacing;
        BigReal MSMPadding;
        Bool MsmSerialOn;

	Bool PMEOn;			//  Flag TRUE -> PME active
	BigReal PMETolerance;		//  Direct space tolerance
        BigReal PMEEwaldCoefficient;    //  From tolerance and cutoff
	int PMEInterpOrder;		//  Order of interpolation
	int PMEGridSizeX;		//  No. of grid points in x dim
	int PMEGridSizeY;		//  No. of grid points in y dim
	int PMEGridSizeZ;		//  No. of grid points in z dim
	BigReal PMEGridSpacing;		//  Maximum spacing between points
	int PMEProcessors;		//  No. of processors to use
	int PMEMinSlices;		//  Min slices per PME slab
	int PMEMinPoints;		//  Min points per PME pencil
	Bool PMEBarrier;		//  Use barrier before sendTrans
	int PMEPencils;			//  Size of pencil grid in each dim

	Bool useDPME;			//  Flag TRUE -> old DPME code
	Bool useOptPME;                 //  Flag TRUE -> use the scalable version of PME
	Bool useManyToMany;             //  Flag TRUE -> use the manytomany optimization of PME. 
	                                //  This flag requres useOptPME to be set.

	Bool FFTWEstimate;
	Bool FFTWPatient;
	Bool FFTWUseWisdom;
	char FFTWWisdomFile[128];
	char *FFTWWisdomString;

        #ifdef OPENATOM_VERSION
        Bool openatom;                  // Flag TRUE -> OpenAtom QM/MM active
        #endif // OPENATOM_VERSION 

	Bool minimizeCGOn;		//  Flag TRUE-> CG minimization active
        Bool minVerbose;                //  Flag TRUE-> print extra minimization data
	BigReal minTinyStep;		//  Minimization parameter
	BigReal minBabyStep;		//  Minimization parameter
	BigReal minLineGoal;		//  Minimization parameter
	Bool minimizeOn;		//  Flag TRUE-> minimization active
	BigReal maximumMove;		//  Maximum movement per timestep 
					//  during minimization

	Bool sphericalBCOn;		//  Flag TRUE-> spherical boundary 
					//  conditions are active
	zVector sphericalCenter;		//  Center specified by user
	BigReal sphericalBCk1;		//  First force constant for 
					//  spherical BC
	BigReal sphericalBCk2;		//  Second force constant for 
					//  spherical BC
	BigReal sphericalBCr1;		//  First radius for spherical BC
	BigReal sphericalBCr2;		//  Second radius for spherical BC
	int sphericalBCexp1;		//  First radius for spherical BC
	int sphericalBCexp2;		//  Second radius for spherical BC

        Bool cylindricalBCOn;           //  Flag TRUE->cylindrical boundary
                                        //  conditions are active
        zVector cylindricalCenter;
	char cylindricalBCAxis;		//  'x', 'y', or 'z'
        BigReal cylindricalBCr1;
        BigReal cylindricalBCr2;
        BigReal cylindricalBCl1;
        BigReal cylindricalBCl2;
        int cylindricalBCexp1;
        int cylindricalBCexp2;
        BigReal cylindricalBCk1;
        BigReal cylindricalBCk2;

	Bool eFieldOn;                  //  Should a electric field be applied
	zVector eField;                 //  Electric field vector to be applied
	BigReal eFieldFreq;		// Frequency of the electric field
	BigReal eFieldPhase;		// Phase phi, cos(w*t-phi*PI/180) 

	Bool stirOn;                   // Should a stirring torque be applied
	char stirFilename[128];	       // Stirring filename (atoms marked)
	//do the below two even needed to be defined?
	BigReal stirStartingTheta;     // Stir starting theta offset
	BigReal stirVel;               // Stir angular velocity
	BigReal stirK;                 // Stir force harmonic spring constant
        zVector stirAxis;              // Direction of stir axis
        zVector stirPivot;             // Pivot point of stir axis

	Bool extraBondsOn;		// read extra bonded forces

	Bool consForceOn;		//  Should constant force be applied
  char consForceFile[128];
	BigReal consForceScaling;

	int outputEnergies;		//  Number of timesteps between energy
					//  outputs

	int outputMomenta;		//  Number of timesteps between momentum
					//  outputs

	int outputTiming;		//  Number of timesteps between timing
					//  outputs

	int outputCudaTiming;		//  Number of timesteps between timing
					//  outputs of CUDA code

	int outputPressure;		//  Number of timesteps between pressure
					//  tensor outputs

	Bool mergeCrossterms;		//  Merge crossterm energy w/ dihedrals

	int firstTimestep;		//  Starting timestep.  Will be 0 unless
					//  restarting a simulation

	MTSChoices MTSAlgorithm;	//  What multiple timestep algorithm
					//  to use

	int longSplitting;		//  What electrostatic splitting 	
					//  to use

	int splitPatch;			// How are patches determined?
	BigReal hgroupCutoff;		// what is the added hydrogen margin?

	int mollyOn;			// mollify long range forces?
	BigReal mollyTol;		// error tolerance for molly
	int mollyIter;			// max number of iterations for molly

        int rigidBonds;                 // what type of rigid bonds to hydrogens
                                        // none, all, or only water

        BigReal rigidTol;               // error tolerance for rigid bonds
        int rigidIter;                  // Number of NR iterations 
	int rigidDie;			// die if rigidTol not achieved

	Bool useSettle;			// Use SETTLE; requires rigid waters

	Bool testOn;			//  Do tests rather than simulation
	Bool commOnly;			//  Don't do any force evaluations

	int totalAtoms;			//  Total Number of atoms in simulation
        int maxSelfPart;                // maximum number of self partitions
                                        // that a patch can be split into
        int maxPairPart;                // maximum number of pair partitions
                                        // that a patch can be split into
        int numAtomsSelf;               // maximum number of atoms in a single
                                        // self-compute 
        int numAtomsSelf2;              // maximum number of atoms in a pair compute
                                        // in the presence of twoAwayX,Y,Z options
        int numAtomsPair;               // maximum number of atoms in a single
                                        // pair-compute 
        int numAtomsPair2;              // maximum number of atoms in a single
                                        // pair-compute 
        int minAtomsPerPatch;           // minimum average atoms per patch
                                        //  (may create larger patches) 
	int maxExclusionFlags;		// maximum size of exclusion check list
					// for any given atom
	Bool outputPatchDetails;	// print number of atoms per patch
        Bool staticAtomAssignment;      // never migrate atoms

	//
        // hydrogen bond simulation parameters
        //

        // should the hydrogen bond term be used?  If FALSE, all other
	// hydrogen bond parameters are unnecessary in simulation.
	Bool HydrogenBonds;

	// should the antecedent atom be used in the calculation of hbonds?
	Bool useAntecedent;

	// exponents used in hydrogen bond energy function:
	//   aaAngleExp = exp for H-A-AA angle term (n)
	//   haAngleExp = exp for D-H-A angle term (m)
	//   distAttExp = exp for attractive A-D distance term (j)
	//   distRepExp = exp for repulsive A-D distance term (i)
	int aaAngleExp, haAngleExp, distAttExp, distRepExp;

	// cutoff D-H-A angle, and on/off angles for switch fcn (in degrees)
	BigReal dhaCutoffAngle, dhaOnAngle, dhaOffAngle;

	// cutoff distance for D-A separation in hbonds (in Angstroms), and
	// on/off distances for hbond radial term switching function
	BigReal daCutoffDist, daOnDist, daOffDist;

	// IMD parameters
	int IMDon;    // enable IMD
	int IMDport;  // port on which to listen for connections
 	int IMDfreq;  // frequency at which coordinates will be available
        int IMDwait;  // if true, pause the simulation when there is no
                      // connection
 	int IMDignore;  // IMD connection does not influence simulation
                        // only sends coordinates and energies to VMD
        
        // AMBER options
        Bool amberOn; // FLAG TRUE-> amber force field is used
        Bool readExclusions; // FLAG TRUE-> Read exclusions from parm file
        BigReal vdwscale14; //  Scaling factor for 1-4 VDW interactions

	// GROMACS options
	Bool gromacsOn; // FLAG TRUE -> gromacs-style force field is used

	// OPLS options
	Bool vdwGeometricSigma;  // Lennard-J sigma uses geometric mean

	// ScriptTcl argument passing
	BigReal scriptArg1;
	BigReal scriptArg2;
	BigReal scriptArg3;
	BigReal scriptArg4;
	BigReal scriptArg5;

	Bool useCompressedPsf;
	Bool genCompressedPsf;

	Bool usePluginIO;

        Bool mallocTest;
        Bool printExclusions;

	//default value is -1
	int proxySendSpanningTree;
	int proxyRecvSpanningTree;

    int proxyTreeBranchFactor;


    //fields needed for Parallel IO Input
    int numinputprocs;
    char *binAtomFile;
    char *binCoorFile;
    char *binVelFile; 
    
    //fields needed for Parallel IO Output
    int numoutputprocs; 
    int numoutputwrts;

	char computeMapFilename[128];		//  store compute map
        Bool storeComputeMap;
        Bool loadComputeMap;

public:

        SimParameters() : mgridforcelist() {};
        SimParameters(ConfigList *c, char *&cwd) : mgridforcelist() {
	  initialize_config_data(c,cwd);
	};
        ~SimParameters() {};

	void initialize_config_data(ConfigList *, char *&cwd);
					//  Initialize SimParameters data
					//  from the ConfigList object
	void send_SimParameters(MOStream *);	
					//  Used by the master process
					//  to send the paramters to
					//  the other processors
	void receive_SimParameters(MIStream *);  
					//  Used by the other processors
					//  to receive the data from the
					//  master process
	void scriptSet(const char *, const char *);
					//  Set parameters at run time

	int isSendSpanningTreeOn(){ return proxySendSpanningTree == 1; }
	int isSendSpanningTreeUnset() { return proxySendSpanningTree == -1; }
	int isRecvSpanningTreeOn(){ return proxyRecvSpanningTree == 1; }
	int isRecvSpanningTreeUnset() { return proxyRecvSpanningTree == -1; }

private:

	void config_parser(ParseOptions &opts);

	void config_parser_basic(ParseOptions &opts);
	void config_parser_fileio(ParseOptions &opts);
	void config_parser_fullelect(ParseOptions &opts);
	void config_parser_methods(ParseOptions &opts);
	void config_parser_constraints(ParseOptions &opts);
#ifdef OPENATOM_VERSION
	void config_parser_openatom(ParseOptions &opts);
#endif //OPENATOM_VERSION
	/* BEGIN gf */
	void config_parser_gridforce(ParseOptions &opts);
	/* END gf */
	void config_parser_movdrag(ParseOptions &opts);
	void config_parser_rotdrag(ParseOptions &opts);
	void config_parser_constorque(ParseOptions &opts);
	void config_parser_boundary(ParseOptions &opts);
	void config_parser_misc(ParseOptions &opts);
   	void config_parser_mgridforce(ParseOptions &opts);
     	void parse_mgrid_string_param(ConfigList *config,
     	                              const char *fieldname, char** dest);
     	void parse_mgrid_params(ConfigList *config);
       	void print_mgrid_params();
  
	void check_config(ParseOptions &opts, ConfigList *config, char *&cwd);

	void print_config(ParseOptions &opts, ConfigList *config, char *&cwd);

	void create_output_directories(const char *dirname);

	int fmaFrequency;		//  outdated parameter name
	char loadBalancer[64];		//  Load balancer
	char loadStrategy[64];		//  Load balancing strategy

};

#endif

